@@ -1,32 +1,26 @@
-"""Definition of a single parameter"""
+# pylint: disable=too-many-lines
+"""Definition of a single parameter
+
+Attributes:
+    PARAM_MAPPINGS: The type to Param mappings, used for params to init a
+        parameter. Managed by `register_param`
+"""
 import ast
 import json
-from typing import (
-    Optional,
-    List,
-    Any,
-    Callable,
-    Type,
-    Dict,
-    Set,
-    Tuple,
-    Union
-)
+from typing import (Optional, List, Any, Callable, Type,
+                    Dict, Set, Tuple, Union, Iterator)
 from itertools import product
 from pathlib import Path
 from diot import OrderedDiot
 from .utils import cast_to, parse_type, logger, TYPE_NAMES, Namespace
 from .defaults import POSITIONAL
-from .exceptions import (
-    PyParamException,
-    PyParamValueError,
-    PyParamTypeError,
-    PyParamNameError,
-)
+from .completer import CompleterParam
+from .exceptions import (PyParamException, PyParamValueError,
+                         PyParamTypeError, PyParamNameError)
 
 PARAM_MAPPINGS: Dict[str, Type['Param']] = {}
 
-class Param:
+class Param(CompleterParam):
     """Base class for parameter
 
     Args:
@@ -43,6 +37,8 @@ class Param:
         callback: The callback to modify the final value
         argname_shorten: Whether show shortened name for parameters
             under namespace parameters
+        complete_callback: The callback for complete the values of the
+            parameter
         **kwargs: Additional keyword arguments
 
     Attributes:
@@ -60,6 +56,8 @@ class Param:
         hit: Whether the parameter is just hit
         ns_param: The namespace parameter where this parameter
             is under
+        complete_callback: The callback for complete the values of the
+            parameter
         is_help: Whether this is a help parameter
         _desc: The raw description of the parameter
         _stack: The stack to push the values
@@ -67,7 +65,6 @@ class Param:
         _kwargs: other kwargs
     """
     # pylint: disable=too-many-instance-attributes
-
     type: Optional[str] = None
     type_aliases: List[str] = []
 
@@ -75,7 +72,7 @@ class Param:
     def on_register(cls):
         """Opens opportunity to do something when a parameter is registered"""
 
-    def __init__(self, # pylint: disable=too-many-arguments
+    def __init__(self,
                  names: List[str],
                  default: Any,
                  desc: Optional[List[str]],
@@ -85,11 +82,12 @@ class Param:
                  subtype: Optional[bool] = None,
                  type_frozen: bool = True,
                  callback: Optional[Callable] = None,
+                 complete_callback: Optional[Callable] = None,
                  argname_shorten: bool = True,
                  **kwargs: Dict[str, Any]):
+        # pylint: disable=too-many-arguments
         """Constructor"""
         self.names: List[str] = names
-
         self.default: Any = default
         self.prefix: str = prefix
         self.show: bool = show
@@ -97,6 +95,7 @@ class Param:
         self.subtype: Optional[str] = subtype
         self.type_frozen: bool = type_frozen
         self.callback: Optional[Callable] = callback
+        self.complete_callback: Optional[Callable] = complete_callback
         self.argname_shorten: bool = argname_shorten
         self.hit: bool = False
         self.is_help: bool = False
@@ -122,11 +121,9 @@ class Param:
                 raise PyParamNameError(
                     "Parameter names must have the same number of namespaces."
                 )
-            if not namespaces:
-                namespaces = [{part} for part in parts[:-1]]
-            else:
-                for i, part in enumerate(parts[:-1]):
-                    namespaces[i].add(part)
+            namespaces = namespaces or [{part} for part in parts[:-1]]
+            for i, part in enumerate(parts[:-1]):
+                namespaces[i].add(part)
             terminals.add(parts[-1])
         return [list(ns) for ns in namespaces], list(terminals)
 
@@ -209,15 +206,10 @@ class Param:
             return self
 
         if self.type_frozen:
-            raise PyParamTypeError(
-                f"Type of argument {self.namestr()!r} "
-                "is not overwritable"
-            )
-
+            raise PyParamTypeError(f"Type of argument {self.namestr()!r} "
+                                   "is not overwritable")
         logger.warning("Type changed from %r to %r for argument %r",
-                       self.typestr(),
-                       param_type,
-                       self.namestr())
+                       self.typestr(), param_type, self.namestr())
         return self.to(param_type)
 
     def consume(self, value: Any) -> bool:
@@ -238,11 +230,8 @@ class Param:
     def desc(self) -> List[str]:
         """The formatted description using attributes and _kwargs"""
         format_data: dict = {
-            key: val
-            for key, val in self.__dict__.items()
-            if (not key.startswith('_') and
-                key != 'desc' and
-                not callable(val))
+            key: val for key, val in self.__dict__.items()
+            if (not key.startswith('_') and key != 'desc' and not callable(val))
         }
         format_data.update(self._kwargs)
         ret: List[str] = []
@@ -277,16 +266,11 @@ class Param:
         Returns:
             The shortest/longest name of the parameter
         """
-        name: str = list(sorted(self.names, key=len))[
-            0 if 'short' in which else -1
-        ]
-        if not with_prefix:
-            return name
-        return self._prefix_name(name)
+        name: str = list(sorted(self.names,
+                                key=len))[0 if 'short' in which else -1]
+        return name if not with_prefix else self._prefix_name(name)
 
-    def namestr(self,
-                sep: str = ", ",
-                with_prefix: bool = True) -> str:
+    def namestr(self, sep: str = ", ", with_prefix: bool = True) -> str:
         """Get all names connected with a separator.
 
         Args:
@@ -296,10 +280,8 @@ class Param:
             the connected names
         """
         names: list = ['POSITIONAL' if name == POSITIONAL
-                       else self._prefix_name(name)
-                       if with_prefix
-                       else name
-                       for name in sorted(self.names, key=len)]
+                       else self._prefix_name(name) if with_prefix
+                       else name for name in sorted(self.names, key=len)]
         return sep.join(names)
 
     def typestr(self) -> str:
@@ -308,9 +290,7 @@ class Param:
         Returns:
             the string representation of the type
         """
-        if not self.subtype:
-            return self.type
-        return f"{self.type}:{self.subtype}"
+        return self.type if not self.subtype else f"{self.type}:{self.subtype}"
 
     def usagestr(self) -> str:
         """Get the string representation of the parameter in the default usage
@@ -335,10 +315,9 @@ class Param:
         typestr: str = (self.typestr().upper()
                         if self.type_frozen else self.typestr())
         if not self.ns_param or not self.argname_shorten:
-            if self.is_help:
-                return self.namestr()
-            # * makes sure it's not wrapped'
-            return f"{self.namestr()}*<{typestr}>"
+            # * makes sure it's not wrapped
+            return (self.namestr() if self.is_help
+                    else f"{self.namestr()}*<{typestr}>")
 
         ret: List[str] = []
         for term in sorted(self.terminals, key=len):
@@ -365,11 +344,33 @@ class Param:
             show=self.show,
             required=self.required,
             subtype=sub_type,
-            callback=self.callback
+            callback=self.callback,
+            complete_callback=self.complete_callback,
+            argname_shorten=self.argname_shorten
         )
         param.ns_param = self.ns_param
         return param
 
+    def copy(self) -> Type['Param']:
+        """Copy a parameter so that it can be reused.
+
+        Returns:
+            The copy of the parameter
+        """
+        return self.__class__(
+            names=self.names,
+            default=self.default,
+            desc=self.desc,
+            prefix=self.prefix,
+            show=self.show,
+            required=self.required,
+            subtype=self.subtype,
+            callback=self.callback,
+            type_frozen=self.type_frozen,
+            complete_callback=self.complete_callback,
+            argname_shorten=self.argname_shorten
+        )
+
     @property
     def default_group(self) -> str:
         """Get the default group of the parameter
@@ -378,9 +379,8 @@ class Param:
             the default group name
         """
         ret: str = "REQUIRED OPTIONS" if self.required else "OPTIONAL OPTIONS"
-        if not self.ns_param:
-            return ret
-        return f"{ret} UNDER {self.ns_param.name('long')}"
+        return (ret if not self.ns_param
+                else f"{ret} UNDER {self.ns_param.name('long')}")
 
     @property
     def desc_with_default(self) -> Optional[List[str]]:
@@ -393,22 +393,18 @@ class Param:
         if self.is_help:
             return self.desc
 
-        if (
-                self.required or (
-                    self.desc and
-                    any("Default:" in desc or "DEFAULT:" in desc
-                        for desc in self.desc)
-                )
-        ):
+        if (self.required or (
+                self.desc and
+                any("Default:" in desc or "DEFAULT:" in desc
+                    for desc in self.desc)
+        )):
             return None if self.desc is None else self.desc[:]
 
         desc: List[str] = self.desc[:] if self.desc else ['']
-
         if desc[0] and not desc[0][-1:].isspace():
             desc[0] += " "
 
         default_str: str = str(self.default) or "''"
-
         desc[0] += f"Default: {default_str}"
         return desc
 
@@ -440,7 +436,6 @@ class Param:
         Returns:
             The parsed value of thie parameter
         """
-
         if not self._stack:
             if self.required:
                 raise PyParamValueError("Argument is required.")
@@ -462,7 +457,6 @@ class Param:
         self._value_cached = self._value()
         return self._value_cached
 
-
     def apply_callback(self, all_values: Namespace) -> Any:
         """Apply the callback function to the value
 
@@ -477,7 +471,6 @@ class Param:
         """
         if not callable(self.callback):
             return self.value
-
         try:
             try:
                 val: Any = self.callback(self.value, all_values)
@@ -492,7 +485,6 @@ class Param:
 
 class ParamAuto(Param):
     """An auto parameter whose value is automatically casted"""
-
     type: str = 'auto'
 
     def _value(self) -> Any:
@@ -591,8 +583,7 @@ class ParamBool(Param):
         try:
             cast_to(value, 'bool')
         except (PyParamValueError, PyParamTypeError):
-            # cannot cast, don't consume
-            return False
+            return False # cannot cast, don't consume
         else:
             self.push(value)
             return True
@@ -605,6 +596,33 @@ class ParamBool(Param):
         self._stack = []
         return ret
 
+    def complete_value(
+            self,
+            current: str,
+            prefix: str = ''
+    ) -> Optional[Union[
+        str, Iterator[Tuple[str]],
+        Iterator[Tuple[str, str]], Iterator[Tuple[str, str, str]]
+    ]]:
+        """Get the completion candidates for the current parameter"""
+        if self.is_help:
+            return ''
+        if callable(self.complete_callback):
+            return super().complete_value(current, prefix)
+        if current:
+            trues: List[str] = ('True', 'true', 'TRUE', '1')
+            falses: List[str] = ('False', 'false', 'FALSE', '0')
+            ret: List[Tuple[str, str, str]] = []
+            for cand in trues:
+                if cand.startswith(current):
+                    ret.append((f'{prefix}{cand}', 'plain', 'Value True'))
+
+            for cand in falses:
+                if cand.startswith(current):
+                    ret.append((f'{prefix}{cand}', 'plain', 'Value False'))
+            return ret
+        return ''
+
 class ParamCount(Param):
     """A bool parameter whose value is automatically casted into a bool"""
     type: str = 'count'
@@ -623,10 +641,8 @@ class ParamCount(Param):
                 "Count argument must have a short name."
             )
 
-        if 'max' in self._kwargs and (
-                not isinstance(self._kwargs['max'], int) or
-                self._kwargs['max'] <= 0
-        ):
+        self._kwargs.setdefault('max', 0) # 0: no max
+        if not isinstance(self._kwargs['max'], int) or self._kwargs['max'] < 0:
             raise PyParamValueError(
                 "Argument 'max' for count argument must be a positive integer"
             )
@@ -637,13 +653,9 @@ class ParamCount(Param):
 
     def consume(self, value: str) -> bool:
         """Should I consume given parameter?"""
-        if not self.hit:
-            return False
-
-        if value.isdigit():
+        if self.hit and value.isdigit():
             self.push(value)
             return True
-
         return False
 
     def _value(self) -> int:
@@ -662,30 +674,96 @@ class ParamCount(Param):
                     retval = len(val) + 1
                     break
         if retval is None:
-            raise PyParamValueError(
-                "Expect repeated short names or an integer "
-                "as count argument value."
-            )
-        if ('max' in self._kwargs and
-                self._kwargs['max'] and
-                retval > self._kwargs['max']):
-            raise PyParamValueError(
-                f"{retval} is greater than the max of "
-                f"{self._kwargs['max']}."
-            )
-
+            raise PyParamValueError("Expect repeated short names or "
+                                    "an integer as count argument value.")
+        if self._kwargs['max'] and retval > self._kwargs['max']:
+            raise PyParamValueError(f"{retval} is greater than "
+                                    f"the max of {self._kwargs['max']}.")
         return retval
 
+    def complete_name(self, current: str) -> Iterator[Tuple[str, str]]:
+        """Complete names for a count parameter
+
+        Since we have -v, -vv, -vvv allowed for a count parameter, we need
+        to put them in the completions, too.
+        """
+        # check if current is trying to do so
+        for name in self.names:
+            if len(name) != 1:
+                continue
+            # current is not like -vvv
+            if self._prefix_name(name) + len(current[2:]) * name != current:
+                continue
+            # check the max
+            value: int = len(current) - 1
+            if self._kwargs['max'] and value >= self._kwargs['max']:
+                # already max'ed, no further completion
+                continue
+
+            ncompletes: int = (min(self._kwargs['max'] - value, 2)
+                               if self._kwargs['max'] else 2)
+            for i in range(ncompletes):
+                yield current + name * (i + 1), self.desc[0].splitlines()[0]
+            break
+
+        else:
+            yield from super().complete_name(current)
+
 class ParamPath(Param):
     """A path parameter whose value is automatically casted into a pathlib.Path
     """
     type: str = 'path'
-    type_aliases: List[str] = ['p']
+    type_aliases: List[str] = ['p', 'file']
 
     def _value(self) -> Path:
         val: Optional[Path] = super()._value()
         return None if val is None else Path(val)
 
+    def complete_value(
+            self,
+            current: str,
+            prefix: str = ''
+    ) -> Optional[Union[
+        str, Iterator[Tuple[str]],
+        Iterator[Tuple[str, str]], Iterator[Tuple[str, str, str]]
+    ]]:
+        """Generate file paths with given current prefix
+        as completion candidates
+
+        Args:
+            current: The current word or prefix under cursor
+        """
+        if callable(self.complete_callback):
+            return super().complete_value(current, prefix)
+        return [(current, 'file', prefix)]
+
+class ParamDir(ParamPath):
+    """Subclass of ParamPath.
+
+    It does not make any difference with pyparam. However, it works differently
+    for completions. The completion items for this param will only give
+    directories instead of all paths
+    """
+    type: str = 'dir'
+    type_aliases: List[str] = []
+
+    def complete_value(
+            self,
+            current: str,
+            prefix: str = ''
+    ) -> Optional[Union[
+        str, Iterator[Tuple[str]],
+        Iterator[Tuple[str, str]], Iterator[Tuple[str, str, str]]
+    ]]:
+        """Generate dir paths with given current prefix as completion candidates
+
+        Args:
+            current: The current word or prefix under cursor
+        """
+        if callable(self.complete_callback):
+            return super().complete_value(current, prefix)
+        return [(current, 'dir', prefix)]
+
 class ParamPy(Param):
     """A parameter whose value will be ast.literal_eval'ed"""
     type: str = 'py'
@@ -706,6 +784,7 @@ class ParamList(Param):
     """A parameter whose value is a list"""
     type: str = 'list'
     type_aliases: List[str] = ['l', 'a', 'array']
+    complete_relapse: bool = True
 
     @classmethod
     def on_register(cls):
@@ -727,7 +806,6 @@ class ParamList(Param):
         if param_type == 'reset':
             self._stack = []
             return self
-
         return super().overwrite_type(param_type)
 
     def consume(self, value: str) -> bool:
@@ -744,11 +822,11 @@ class ParamList(Param):
 
     def _value(self) -> List[Any]:
         """Get the value a list parameter"""
-        ret = [
-            cast_to(val, self.subtype)
-            for sublist in self._stack
-            for val in sublist
-        ]
+        ret = [cast_to(val, self.subtype) for sublist in self._stack
+               for val in sublist]
+
+        if self.required and not ret:
+            raise PyParamValueError("Argument is required.")
         self._stack = []
         return ret
 
@@ -761,29 +839,48 @@ class ParamChoice(Param):
         super().__init__(*args, **kwargs)
 
         if 'choices' not in self._kwargs:
-            raise PyParamValueError(
-                "Argument 'choices' is required for ParamChoice."
-            )
+            raise PyParamValueError("Argument 'choices' is required "
+                                    "for ParamChoice.")
 
         if not isinstance(self._kwargs['choices'], (list, tuple)):
-            raise PyParamValueError(
-                "Argument 'choices' must be a list or a tuple."
-            )
+            raise PyParamValueError("Argument 'choices' must be "
+                                    "a list or a tuple.")
+
+        if not self.required and self.default is None:
+            self.default = self._kwargs['choices'][0]
 
     def _value(self) -> Any:
         val = super()._value()
 
         if val not in self._kwargs['choices']:
-            if val is None:
-                val = self._kwargs['choices'][0]
-            else:
-                raise PyParamValueError(
-                    f"{val} is not one of {self._kwargs['choices']}"
-                )
-
+            raise PyParamValueError(
+                f"{val} is not one of {self._kwargs['choices']}"
+            )
         self._stack = []
         return val
 
+    def complete_value(
+            self,
+            current: str,
+            prefix: str = ''
+    ) -> Optional[Union[
+        str, Iterator[Tuple[str]],
+        Iterator[Tuple[str, str]], Iterator[Tuple[str, str, str]]
+    ]]:
+        """Generate choices with given current prefix as completion candidates
+
+        Args:
+            current: The current word or prefix under cursor
+        """
+        if callable(self.complete_callback):
+            return super().complete_value(current, prefix)
+
+        ret: Iterator[Tuple[str]] = []
+        for choice in self._kwargs['choices']:
+            if choice.startswith(current):
+                ret.append((f"{prefix}{choice}", ))
+        return ret
+
 class ParamNamespace(Param):
     """A pseudo parameter serving as a namespace for parameters under it
 
@@ -802,26 +899,34 @@ class ParamNamespace(Param):
         ]
         kwargs['default'] = None
         super().__init__(*args, **kwargs)
-        # for my params
-        self._stack = OrderedDiot()
+        self._stack = OrderedDiot() # for my decendents
 
     @property
     def default_group(self) -> str:
         """Get the default group of the parameter"""
-        if any(param.required for param in self._stack.values()):
-            ret: str = 'REQUIRED'
-        else:
-            ret: str = 'OPTIONAL'
-        if not self.ns_param:
-            return f'{ret} OPTIONS'
-        return (f'{ret} OPTIONS UNDER '
-                f'{self.ns_param.name("long")}')
+        ret: str = ('REQUIRED'
+                    if any(param.required for param in self._stack.values())
+                    else 'OPTIONAL')
+        return (f'{ret} OPTIONS' if not self.ns_param
+                else f'{ret} OPTIONS UNDER {self.ns_param.name("long")}')
 
     @property
     def desc_with_default(self) -> Optional[List[str]]:
         """Namespace parameters do not have a default value"""
         return self.desc[:]
 
+    @property
+    def decendents(self) -> List[Type['Param']]:
+        """Get all decendents of this namespace parameter"""
+        ret: List[Type['Param']] = []
+        ret_append: Callable = ret.append
+        for param in self._stack.values():
+            if param not in ret:
+                ret_append(param)
+            if isinstance(param, ParamNamespace):
+                ret.extend(param.decendents)
+        return ret
+
     def consume(self, value: str) -> bool:
         """Should I consume given parameter?"""
         return False
@@ -896,9 +1001,19 @@ class ParamNamespace(Param):
             if param_name not in val:
                 for name in param.terminals:
                     val[name] = param.value
-
         return val
 
+    def copy(self) -> Type['Param']:
+        """Copy a parameter so that it can be reused.
+
+        Returns:
+            The copy of the parameter
+        """
+        copied = super().copy()
+        copied._stack = OrderedDiot([(key, param.copy())
+                                     for key, param in self._stack.items()])
+        return copied
+
     def apply_callback(self, all_values: Namespace) -> Any:
         ns_callback_applied = Namespace()
         for param_name, param in self._stack.items():
@@ -931,23 +1046,14 @@ def register_param(param: Param) -> None:
     """
     for alias in param.type_aliases + [param.type]:
         if alias in TYPE_NAMES:
-            raise PyParamNameError(
-                f'Type name has already been registered: {alias}'
-            )
+            raise PyParamNameError('Type name has already been '
+                                   f'registered: {alias}')
         TYPE_NAMES[alias] = param.type
 
     PARAM_MAPPINGS[param.type] = param
     param.on_register()
 
-register_param(ParamAuto)
-register_param(ParamInt)
-register_param(ParamStr)
-register_param(ParamFloat)
-register_param(ParamBool)
-register_param(ParamCount)
-register_param(ParamPath)
-register_param(ParamPy)
-register_param(ParamJson)
-register_param(ParamList)
-register_param(ParamChoice)
-register_param(ParamNamespace)
+for param_class in (ParamAuto, ParamInt, ParamStr, ParamFloat, ParamBool,
+                    ParamCount, ParamPath, ParamDir, ParamPy, ParamJson,
+                    ParamList, ParamChoice, ParamNamespace):
+    register_param(param_class)
