@@ -25,10 +25,11 @@ from .utils import (
 from .defaults import POSITIONAL
 from .param import PARAM_MAPPINGS
 from .help import HelpAssembler
+from .completer import Completer
 from .exceptions import PyParamNameError
 
 
-class Params: # pylint: disable=too-many-instance-attributes
+class Params(Completer): # pylint: disable=too-many-instance-attributes
     """Params, served as root params or subcommands
 
     Args:
@@ -77,7 +78,7 @@ class Params: # pylint: disable=too-many-instance-attributes
                  names: Optional[Union[str, List[str]]] = None,
                  desc: Optional[Union[List[str], str]] = 'No description',
                  prog: Optional[str] = None,
-                 help_keys: Union[str, List[str]] = 'h,help,H',
+                 help_keys: Union[str, List[str]] = 'h,help',
                  help_cmds: Union[str, List[str]] = 'help',
                  help_on_void: bool = True,
                  help_callback: Optional[Callable] = None,
@@ -109,6 +110,7 @@ class Params: # pylint: disable=too-many-instance-attributes
         self.assembler: HelpAssembler = HelpAssembler(
             self.prog, theme, help_callback
         )
+        super().__init__()
 
     def name(self, which: str = 'short') -> str:
         """Get the shortest/longest name of the parameter
@@ -185,7 +187,7 @@ class Params: # pylint: disable=too-many-instance-attributes
                 self.params[name] = param
 
     def add_param(self, # pylint: disable=too-many-arguments
-                  names: Union[str, List[str]],
+                  names: Union[str, List[str], Type["Param"]],
                   default: Any = None,
                   # pylint: disable=redefined-builtin
                   type: Optional[str] = None,
@@ -197,11 +199,17 @@ class Params: # pylint: disable=too-many-instance-attributes
                   force: bool = False,
                   type_fronzen: bool = True,
                   argname_shorten: bool = True,
+                  complete_callback: Optional[Callable] = None,
                   **kwargs) -> Type["Param"]:
         """Add an argument
 
         Args:
-            names: names of the argument
+            names: names of the argument or a parameter defined somewhere else
+                For example, in case we want to reuse a parameter
+                >>> param = cmd1.add_param('n,name')
+                >>> # reuse it:
+                >>> cmd2.add_param(param)
+                >>> # other arguments will be ignored, except force
             default: The default value for the argument
             type: The type of the argument
                 Including single value type and complex one
@@ -225,6 +233,8 @@ class Params: # pylint: disable=too-many-instance-attributes
                 the commone line
             argname_shorten: Whether show shortened name for parameter
                 under namespace parameter
+            complete_callback: The callback for complete the values of the
+                parameter
             **kwargs: Additional keyword arguments
 
         Raises:
@@ -234,33 +244,39 @@ class Params: # pylint: disable=too-many-instance-attributes
             Param: The added parameter
         """
         # pylint: disable=too-many-locals
-        names: List[str] = always_list(names)
-        if type is None:
-            if POSITIONAL in names and default is None:
-                type = 'list'
-            else:
-                type = type_from_value(default)
+        if isinstance(names, (str, list)):
+            names: List[str] = always_list(names)
+            if type is None:
+                if POSITIONAL in names and default is None:
+                    type = 'list'
+                else:
+                    type = type_from_value(default)
 
-        # Type: Optional[str], Optional[str]
-        maintype, subtype = parse_type(type.__name__
-                                       if callable(type)
-                                       else type)
+            # Type: Optional[str], Optional[str]
+            maintype, subtype = parse_type(type.__name__
+                                           if callable(type)
+                                           else type)
 
-        param: Type['Param'] = PARAM_MAPPINGS[maintype](
-            names=names,
-            default=default,
-            desc=None if desc is None else always_list(desc, strip=False,
-                                                       split=False),
-            prefix=self.prefix,
-            show=show,
-            required=required,
-            subtype=subtype,
-            type_fronzen=type_fronzen,
-            callback=callback,
-            argname_shorten=argname_shorten,
-            **kwargs
-        )
-        if any(name in self.help_keys for name in names):
+            param: Type['Param'] = PARAM_MAPPINGS[maintype](
+                names=names,
+                default=default,
+                desc=None if desc is None else always_list(desc,
+                                                           strip=False,
+                                                           split=False),
+                prefix=self.prefix,
+                show=show,
+                required=required,
+                subtype=subtype,
+                type_fronzen=type_fronzen,
+                callback=callback,
+                argname_shorten=argname_shorten,
+                complete_callback=complete_callback,
+                **kwargs
+            )
+        else:
+            param = names.copy()
+
+        if any(name in self.help_keys for name in param.names):
             param.is_help = True
 
         if param.namespaces(0):
@@ -272,7 +288,7 @@ class Params: # pylint: disable=too-many-instance-attributes
             ns_param.push(param)
 
         else:
-            for name in names:
+            for name in param.names:
                 # check if parameter has been added
                 if not force and (name in self.params or name in self.commands):
                     raise PyParamNameError(
@@ -445,7 +461,12 @@ class Params: # pylint: disable=too-many-instance-attributes
                 desc="Command name to print help for"
             )
 
-        args = sys.argv[1:] if args is None else args
+        if args is None:
+            # enable completion only when we are trying to parse sys.argv
+            if self.comp_shell:
+                print("\n".join(self.complete()))
+                sys.exit(0)
+            args = sys.argv[1:]
 
         if not args and self.help_on_void:
             self.print_help()
@@ -680,8 +701,6 @@ class Params: # pylint: disable=too-many-instance-attributes
                     name_with_attached, self.prefix, allow_attached=True
                 )
             )
-            print(param_name, param_type, param_value)
-            print(param_name2, param_type2, param_value2)
             # Use them only if we found a param_name2 and
             # arbitrary: not previous param_name found
             # otherwise: parameter with param_name2 exists
@@ -689,7 +708,6 @@ class Params: # pylint: disable=too-many-instance-attributes
                     (self.arbitrary and param_name is None) or
                     self.get_param(param_name2)
             ):
-                print(33)
                 param_name, param_type, param_value = (
                     param_name2, param_type2, param_value2
                 )
